""" HomeKit Flashing Light Accessory """
from __future__ import annotations
from typing import TYPE_CHECKING

import logging
from threading import Timer

from pyhap.characteristic import Characteristic
from pyhap.const import CATEGORY_LIGHTBULB

from carconnectivity.commands import GenericCommand
from carconnectivity.errors import SetterError
from carconnectivity.command_impl import HonkAndFlashCommand

from carconnectivity_plugins.homekit.accessories.generic_accessory import GenericAccessory

if TYPE_CHECKING:
    from typing import Optional, Any, Dict

    from pyhap.service import Service
    from pyhap.accessory_driver import AccessoryDriver

    from carconnectivity.vehicle import GenericVehicle

    from carconnectivity_plugins.homekit.accessories.bridge import CarConnectivityBridge


LOG: logging.Logger = logging.getLogger("carconnectivity.plugins.homekit.charging_plug")


class FlashingLightAccessory(GenericAccessory):
    """Flashing Light Accessory"""
    category: int = CATEGORY_LIGHTBULB

    # pylint: disable-next=too-many-arguments,too-many-positional-arguments
    def __init__(self, driver: AccessoryDriver, bridge: CarConnectivityBridge, aid: int, id_str: str, vin: str, display_name: str,
                 vehicle: GenericVehicle) -> None:
        super().__init__(driver=driver, bridge=bridge, display_name=display_name, aid=aid, vin=vin, id_str=id_str)
        self.vehicle: GenericVehicle = vehicle

        # pyright: ignore[reportArgumentType]
        self.service: Optional[Service] = self.add_preload_service(service='Lightbulb',  # pyright: ignore[reportArgumentType]
                                                                   chars=['Name', 'ConfiguredName',  # pyright: ignore[reportArgumentType]
                                                                          'On', 'StatusFault'])

        self.char_on: Optional[Characteristic] = None

        self.add_name_characteristics()
        self.add_status_fault_characteristic()

        if self.vehicle is not None and self.vehicle.commands is not None and self.vehicle.commands.commands is not None \
                and 'honk-flash' in self.vehicle.commands.commands:
            self.honk_flash_command: GenericCommand = self.vehicle.commands.commands['honk-flash']
            self.char_on = self.service.configure_char('On', setter_callback=self.__on_hk_on_change)

    def __on_hk_on_change(self, value: Any) -> None:
        try:
            if self.honk_flash_command is not None and self.honk_flash_command.enabled:
                if value is True:
                    LOG.info('Start flashing for 10 seconds')
                    command_args: Dict[str, Any] = {}
                    command_args['command'] = HonkAndFlashCommand.Command.FLASH
                    self.honk_flash_command.value = command_args

                    def reset_state():
                        if self.char_on is not None:
                            self.char_on.set_value(False)

                    timer = Timer(10.0, reset_state)
                    timer.start()
                else:
                    LOG.error('Flashing cannot be turned off, please wait for flashing to stop')
        except SetterError as setter_error:
            LOG.error('Error honk and flash: %s', setter_error)
            if self.char_on is not None:
                self.char_on.set_value(False)
            self.set_status_fault(1, timeout=120)



""" DummyAccessory class module."""
from typing import Literal
from pyhap.accessory import Accessory
from pyhap.const import CATEGORY_OTHER


class DummyAccessory(Accessory):
    """
    DummyAccessory class that inherits from Accessory.

    This is a placeholder when bringing up the plugin from stored values.
    If we don't offer these placeholders to HomeKit, it will remove the accessory from the Home app.

    Attributes:
        category (int): The category of the accessory, set to CATEGORY_OTHER.

    Methods:
        __init__(driver, aid, display_name):
            Initializes the DummyAccessory instance with the given driver, aid, and display_name.

        available:
            Property that always returns False, indicating the accessory is not available.
    """

    category: int = CATEGORY_OTHER

    def __init__(self, driver, aid, display_name) -> None:
        super().__init__(driver=driver, display_name=display_name, aid=aid)

    @property
    def available(self) -> Literal[False]:
        return False



""" HomeKit Outside Temperature Accessory """
from __future__ import annotations
from typing import TYPE_CHECKING

import logging

from pyhap.characteristic import Characteristic
from pyhap.const import CATEGORY_SENSOR

from carconnectivity.observable import Observable
from carconnectivity.attributes import TemperatureAttribute, Temperature

from carconnectivity_plugins.homekit.accessories.generic_accessory import GenericAccessory
from carconnectivity_plugins.homekit.accessories.util import VALUE_TO_TEMPERATURE_UNIT, TEMPERATURE_UNIT_TO_VALUE

if TYPE_CHECKING:
    from typing import Optional, Any

    from pyhap.service import Service
    from pyhap.accessory_driver import AccessoryDriver

    from carconnectivity.vehicle import GenericVehicle

    from carconnectivity_plugins.homekit.accessories.bridge import CarConnectivityBridge


LOG: logging.Logger = logging.getLogger("carconnectivity.plugins.homekit.outside_temperature")


class OutsideTemperatureAccessory(GenericAccessory):
    """Outside Temperature Accessory"""
    category: int = CATEGORY_SENSOR

    # pylint: disable=duplicate-code
    # pylint: disable-next=too-many-arguments,too-many-positional-arguments
    def __init__(self, driver: AccessoryDriver, bridge: CarConnectivityBridge, aid: int, id_str: str, vin: str, display_name: str,
                 vehicle: GenericVehicle) -> None:
        super().__init__(driver=driver, bridge=bridge, display_name=display_name, aid=aid, vin=vin, id_str=id_str)
        self.vehicle: GenericVehicle = vehicle

        # pyright: ignore[reportArgumentType]
        self.service: Optional[Service] = self.add_preload_service(service='TemperatureSensor',  # pyright: ignore[reportArgumentType]
                                                                   chars=['Name', 'ConfiguredName',  # pyright: ignore[reportArgumentType]
                                                                          'TemperatureDisplayUnits', 'CurrentTemperature'])

        self.char_current_temperature: Optional[Characteristic] = None
        self.char_temperature_display_units: Optional[Characteristic] = None

        self.outside_temperature_attribute: Optional[TemperatureAttribute] = None

        self.add_name_characteristics()

        temperature_display_unit: Optional[int] = self.bridge.get_config_item(self.id_str, self.vin, 'TemperatureDisplayUnits')
        if temperature_display_unit is None:
            temperature_display_unit = TEMPERATURE_UNIT_TO_VALUE[Temperature.C]
            self.configured_temperature_unit: Temperature = Temperature.C
        else:
            self.configured_temperature_unit: Temperature = VALUE_TO_TEMPERATURE_UNIT[temperature_display_unit]
        if self.service is not None:
            self.char_temperature_display_units = self.service.configure_char('TemperatureDisplayUnits', value=temperature_display_unit,
                                                                              valid_values={"Celsius": 0, "Fahrenheit": 1},
                                                                              setter_callback=self.__on_hk_temperature_display_units_change)

        if self.vehicle is not None and vehicle.outside_temperature is not None and vehicle.outside_temperature.enabled:
            self.outside_temperature_attribute = self.vehicle.outside_temperature
            self.vehicle.outside_temperature.add_observer(self.__on_cc_outside_temperature_change, flag=Observable.ObserverEvent.VALUE_CHANGED)
            self.char_current_temperature = self.service.configure_char('CurrentTemperature')
            self.__on_cc_outside_temperature_change(self.vehicle.outside_temperature, Observable.ObserverEvent.VALUE_CHANGED)
    # pylint: disable=duplicate-code

    def __on_cc_outside_temperature_change(self, element: Any, flags: Observable.ObserverEvent) -> None:
        if flags & Observable.ObserverEvent.VALUE_CHANGED and isinstance(element, TemperatureAttribute):
            target_temperature_unit: Temperature = Temperature.C
            if self.char_temperature_display_units is not None:
                target_temperature_unit = VALUE_TO_TEMPERATURE_UNIT[self.char_temperature_display_units.get_value()]
            if self.char_current_temperature is not None and element.enabled and element.value is not None:
                self.char_current_temperature.set_value(element.temperature_in(unit=target_temperature_unit))
            LOG.info('targetTemperature Changed: %f', element.temperature_in(unit=target_temperature_unit))
        else:
            LOG.debug('Unsupported event %s', flags)

    def __on_hk_temperature_display_units_change(self, value: int) -> None:
        if value in VALUE_TO_TEMPERATURE_UNIT:
            if self.char_temperature_display_units is not None:
                self.char_temperature_display_units.set_value(value)
            self.bridge.set_config_item(self.id_str, self.vin, 'TemperatureDisplayUnits', value)
            self.configured_temperature_unit = VALUE_TO_TEMPERATURE_UNIT[value]
            if self.outside_temperature_attribute is not None:
                self.__on_cc_outside_temperature_change(element=self.outside_temperature_attribute, flags=Observable.ObserverEvent.VALUE_CHANGED)
        else:
            LOG.error('Invalid temperature display unit: %d', value)



""" Utility functions and constants for HomeKit accessories. """
from __future__ import annotations
from typing import TYPE_CHECKING

from carconnectivity.units import Temperature

if TYPE_CHECKING:
    from typing import Dict

TEMPERATURE_UNIT_TO_VALUE: Dict[Temperature, int] = {Temperature.C: 0, Temperature.F: 1}
VALUE_TO_TEMPERATURE_UNIT: Dict[int, Temperature] = {0: Temperature.C, 1: Temperature.F}



""" HomeKit Climatization Accessory """
from __future__ import annotations
from typing import TYPE_CHECKING

import logging
import threading
from datetime import datetime, timezone

from pyhap.characteristic import Characteristic
from pyhap.const import CATEGORY_AIR_CONDITIONER

from carconnectivity.errors import SetterError
from carconnectivity.commands import GenericCommand
from carconnectivity.observable import Observable
from carconnectivity.vehicle import ElectricVehicle
from carconnectivity.units import Temperature
from carconnectivity.attributes import TemperatureAttribute
from carconnectivity.climatization import Climatization
from carconnectivity.command_impl import ClimatizationStartStopCommand

from carconnectivity_plugins.homekit.accessories.generic_accessory import BatteryGenericVehicleAccessory
from carconnectivity_plugins.homekit.accessories.util import TEMPERATURE_UNIT_TO_VALUE, VALUE_TO_TEMPERATURE_UNIT

if TYPE_CHECKING:
    from typing import Optional, Any, Dict

    from pyhap.service import Service
    from pyhap.accessory_driver import AccessoryDriver

    from carconnectivity.vehicle import GenericVehicle

    from carconnectivity_plugins.homekit.accessories.bridge import CarConnectivityBridge

LOG: logging.Logger = logging.getLogger("carconnectivity.plugins.homekit.climatization")


class ClimatizationAccessory(BatteryGenericVehicleAccessory):  # pylint: disable=too-many-instance-attributes
    """Climatization Accessory"""

    category: int = CATEGORY_AIR_CONDITIONER

    # pylint: disable=duplicate-code
    # pylint: disable-next=too-many-arguments,too-many-positional-arguments,too-many-branches,too-many-statements
    def __init__(self, driver: AccessoryDriver, bridge: CarConnectivityBridge, aid: int, id_str: str, vin: str, display_name: str,
                 vehicle: GenericVehicle) -> None:
        super().__init__(driver=driver, bridge=bridge, display_name=display_name, aid=aid, vin=vin, id_str=id_str, vehicle=vehicle)

        self.update_remaining_duration_timer: Optional[threading.Timer] = None

        # pyright: ignore[reportArgumentType]
        self.service: Optional[Service] = self.add_preload_service(service='Thermostat',  # pyright: ignore[reportArgumentType]
                                                                   chars=['Name', 'ConfiguredName',  # pyright: ignore[reportArgumentType]
                                                                          'CurrentHeatingCoolingState', 'TargetHeatingCoolingState',
                                                                          'TargetTemperature', 'TemperatureDisplayUnits',
                                                                          'RemainingDuration', 'StatusFault'])

        self.char_target_heating_cooling_state: Optional[Characteristic] = None
        self.char_current_heating_cooling_state: Optional[Characteristic] = None
        self.char_target_temperature: Optional[Characteristic] = None
        self.char_temperature_display_units: Optional[Characteristic] = None
        self.char_remaining_duration: Optional[Characteristic] = None
        self.estimated_date_reached: Optional[datetime] = None

        self.target_temperature_attribute: Optional[TemperatureAttribute] = None
        self.climatization_start_stop_command: Optional[GenericCommand] = None
        temperature_display_unit: Optional[int] = self.bridge.get_config_item(self.id_str, self.vin, 'TemperatureDisplayUnits')
        if temperature_display_unit is None:
            temperature_display_unit = TEMPERATURE_UNIT_TO_VALUE[Temperature.C]
            self.configured_temperature_unit: Temperature = Temperature.C
        else:
            self.configured_temperature_unit: Temperature = VALUE_TO_TEMPERATURE_UNIT[temperature_display_unit]
        if self.service is not None:
            self.char_temperature_display_units = self.service.configure_char('TemperatureDisplayUnits', value=temperature_display_unit,
                                                                              valid_values={"Celsius": 0, "Fahrenheit": 1},
                                                                              setter_callback=self.__on_hk_temperature_display_units_change)
        self.current_target_temperature: Optional[float] = None
        if self.vehicle is not None and self.vehicle.climatization is not None \
                and (target_temperature_attribute := self.vehicle.climatization.settings.target_temperature) is not None:
            self.target_temperature_attribute = target_temperature_attribute
            target_temperature_attribute.add_observer(self.__on_cc_target_temperature_change, flag=Observable.ObserverEvent.VALUE_CHANGED)
            if target_temperature_attribute.enabled:
                self.current_target_temperature = target_temperature_attribute.temperature_in(unit=self.configured_temperature_unit)
            if self.current_target_temperature is None:
                if self.configured_temperature_unit == Temperature.C:
                    self.current_target_temperature = 16
                elif self.configured_temperature_unit == Temperature.F:
                    self.current_target_temperature = 61
                else:
                    LOG.error('Invalid temperature unit: %s', self.configured_temperature_unit)
            self.char_target_temperature = self.service.configure_char('TargetTemperature', value=self.current_target_temperature,
                                                                       setter_callback=self.__on_hk_target_temperature_change)
            self.__update_display_units_properties()

        if self.vehicle is not None and self.vehicle.climatization is not None:
            if self.vehicle.climatization.state is not None:
                if self.vehicle.climatization.state.enabled:
                    self.vehicle.climatization.state.add_observer(self.__on_cc_climatization_state_change, flag=Observable.ObserverEvent.VALUE_CHANGED)
                    self.char_current_heating_cooling_state = self.service.configure_char('CurrentHeatingCoolingState')
                    self.__on_cc_climatization_state_change(self.vehicle.climatization.state, Observable.ObserverEvent.VALUE_CHANGED)

            if self.vehicle.climatization.commands is not None and self.vehicle.climatization.commands.contains_command('start-stop'):
                self.climatization_start_stop_command = self.vehicle.climatization.commands.commands['start-stop']
                self.char_target_heating_cooling_state = self.service.configure_char('TargetHeatingCoolingState',
                                                                                     valid_values={'Auto': 3, 'Off': 0},
                                                                                     setter_callback=self.__on_hk_target_heating_cooling_state_changed)
                self.char_target_heating_cooling_state.allow_invalid_client_values = True
                # call on change again to also set the target state
                self.__on_cc_climatization_state_change(self.vehicle.climatization.state, Observable.ObserverEvent.VALUE_CHANGED)

            if self.vehicle.climatization.estimated_date_reached is not None and self.vehicle.climatization.estimated_date_reached.enabled:
                self.vehicle.climatization.estimated_date_reached.add_observer(self.__on_cc_estimated_date_reached_change,
                                                                               flag=Observable.ObserverEvent.UPDATED_NEW_MEASUREMENT)
                self.char_remaining_duration = self.service.configure_char('RemainingDuration')
                self.__on_cc_estimated_date_reached_change(self.vehicle.climatization.estimated_date_reached,
                                                           Observable.ObserverEvent.UPDATED_NEW_MEASUREMENT)

        self.add_name_characteristics()
        self.add_status_fault_characteristic()

        if self.battery_service is None and isinstance(self.vehicle, ElectricVehicle):
            self.add_soc_characteristic()
        else:
            self.vehicle.add_observer(self.__on_cc_car_type_change, flag=Observable.ObserverEvent.UPDATED)
    # pylint: disable=duplicate-code

    def __on_cc_car_type_change(self, element: Any, flags: Observable.ObserverEvent) -> None:
        if flags & Observable.ObserverEvent.UPDATED:
            if isinstance(element, ElectricVehicle):
                self.add_soc_characteristic()
                self.vehicle.type.remove_observer(self.__on_cc_car_type_change)
        else:
            LOG.debug('Unsupported event %s', flags)

    def __on_cc_target_temperature_change(self, element: Any, flags: Observable.ObserverEvent) -> None:
        if flags & Observable.ObserverEvent.VALUE_CHANGED and isinstance(element, TemperatureAttribute):
            target_temperature_unit: Temperature = Temperature.C
            if self.char_temperature_display_units is not None:
                target_temperature_unit = VALUE_TO_TEMPERATURE_UNIT[self.char_temperature_display_units.get_value()]
            if self.char_target_temperature is not None and element.enabled and element.value is not None:
                self.char_target_temperature.set_value(element.temperature_in(unit=target_temperature_unit))
            LOG.info('targetTemperature Changed: %f', element.temperature_in(unit=target_temperature_unit))
        else:
            LOG.debug('Unsupported event %s', flags)

    def __on_hk_target_temperature_change(self, value: int) -> None:
        if self.char_target_temperature is not None:
            if self.target_temperature_attribute is not None and self.target_temperature_attribute.enabled:
                self.current_target_temperature = value
                if self.configured_temperature_unit is None:
                    unit: Temperature = Temperature.C
                else:
                    unit = self.configured_temperature_unit
                self.target_temperature_attribute.set_value(value=value, unit=unit)
                LOG.info('targetTemperature Changed: %f', value)
            else:
                LOG.error('Target temperature attribute not available')
        else:
            LOG.error('Target temperature characteristic not available')

    def __on_hk_target_heating_cooling_state_changed(self, value: int) -> None:
        try:
            if self.climatization_start_stop_command is not None and self.climatization_start_stop_command.enabled:
                if value in (1, 2, 3):
                    LOG.info('Switch climatization on')
                    command_args: Dict[str, Any] = {}
                    command_args['command'] = ClimatizationStartStopCommand.Command.START
                    self.climatization_start_stop_command.value = command_args
                elif value == 0:
                    LOG.info('Switch climatization off')
                    command_args: Dict[str, Any] = {}
                    command_args['command'] = ClimatizationStartStopCommand.Command.STOP
                    self.climatization_start_stop_command.value = command_args
                else:
                    LOG.error('Input for climatization not understood: %d', value)
        except SetterError as setter_error:
            LOG.error('Error starting climatization: %s', setter_error)
            if self.char_current_heating_cooling_state is not None and self.char_target_heating_cooling_state is not None \
                    and self.char_current_heating_cooling_state.value in [1, 2]:
                self.char_target_heating_cooling_state.set_value(3)
            elif self.char_target_heating_cooling_state is not None:
                self.char_target_heating_cooling_state.set_value(0)
            self.set_status_fault(1, timeout=120)

    def __on_hk_temperature_display_units_change(self, value: int) -> None:
        if value in VALUE_TO_TEMPERATURE_UNIT:
            if self.char_temperature_display_units is not None:
                self.char_temperature_display_units.set_value(value)
            self.bridge.set_config_item(self.id_str, self.vin, 'TemperatureDisplayUnits', value)
            self.configured_temperature_unit = VALUE_TO_TEMPERATURE_UNIT[value]
            self.__update_display_units_properties()
            self.__on_cc_target_temperature_change(element=self.target_temperature_attribute, flags=Observable.ObserverEvent.VALUE_CHANGED)
        else:
            LOG.error('Invalid temperature display unit: %d', value)

    def __update_display_units_properties(self) -> None:
        if self.char_target_temperature is not None and self.char_temperature_display_units is not None:
            if self.configured_temperature_unit == Temperature.C:
                self.char_target_temperature.override_properties(properties={'maxValue': 29.5, 'minStep': 0.5, 'minValue': 16})
            elif self.configured_temperature_unit == Temperature.F:
                self.char_target_temperature.override_properties(properties={'maxValue': 85, 'minStep': 0.5, 'minValue': 61})

    def __on_cc_climatization_state_change(self, element: Any, flags: Observable.ObserverEvent) -> None:  # pylint: disable=too-many-branches
        if flags & Observable.ObserverEvent.VALUE_CHANGED:
            if self.char_current_heating_cooling_state is not None:
                if element.value is None:
                    self.char_current_heating_cooling_state.set_value(0)
                    if self.char_target_heating_cooling_state is not None:
                        self.char_target_heating_cooling_state.set_value(0)
                elif element.value == Climatization.ClimatizationState.HEATING:
                    self.char_current_heating_cooling_state.set_value(1)
                    if self.char_target_heating_cooling_state is not None:
                        self.char_target_heating_cooling_state.set_value(3)
                elif element.value in (Climatization.ClimatizationState.COOLING, Climatization.ClimatizationState.VENTILATION):
                    self.char_current_heating_cooling_state.set_value(2)
                    if self.char_target_heating_cooling_state is not None:
                        self.char_target_heating_cooling_state.set_value(3)
                elif element.value == Climatization.ClimatizationState.OFF:
                    self.char_current_heating_cooling_state.set_value(0)
                    if self.char_target_heating_cooling_state is not None:
                        self.char_target_heating_cooling_state.set_value(0)
                else:
                    self.char_current_heating_cooling_state.set_value(0)
                    if self.char_target_heating_cooling_state is not None:
                        self.char_target_heating_cooling_state.set_value(0)
                    LOG.warning('unsupported climatisationState: %s', element.value.value)
                LOG.debug('Climatization State Changed: %s', element.value.value)
            else:
                LOG.debug('Unsupported event %s', flags)

    def __on_cc_estimated_date_reached_change(self, element: Any, flags: Observable.ObserverEvent) -> None:
        if flags & Observable.ObserverEvent.UPDATED_NEW_MEASUREMENT:
            if self.char_remaining_duration is not None:
                if element.enabled and element.value is not None and isinstance(element.value, datetime):
                    self.estimated_date_reached = element.value
                    self.__update_remaining_duration()
                    LOG.debug('Climatization estimated date reached Changed: %s', self.estimated_date_reached.isoformat())
                else:
                    self.estimated_date_reached = None
                    self.char_remaining_duration.set_value(0)
                    LOG.debug('Climatization estimated date reached Changed: None')
        else:
            LOG.debug('Unsupported event %s', flags)

    # pylint: disable=duplicate-code
    def __update_remaining_duration(self) -> None:
        if self.update_remaining_duration_timer is not None:
            self.update_remaining_duration_timer.cancel()
            self.update_remaining_duration_timer = None
        if self.char_remaining_duration is not None:
            remaining_duration: int = 0
            utc_now: datetime = datetime.now(tz=timezone.utc)
            if self.estimated_date_reached is not None and self.estimated_date_reached > utc_now:
                remaining_duration = round((self.estimated_date_reached - utc_now).total_seconds())
            self.char_remaining_duration.set_value(remaining_duration)
            if remaining_duration > 0 and not self.driver.stop_event.is_set():
                self.update_remaining_duration_timer = threading.Timer(interval=5.0, function=self.__update_remaining_duration)
                self.update_remaining_duration_timer.start()
    # pylint: enable=duplicate-code



""" HomeKit Charging Plug Accessory """
from __future__ import annotations
from typing import TYPE_CHECKING

import logging

from pyhap.characteristic import Characteristic
from pyhap.const import CATEGORY_SENSOR

from carconnectivity.charging_connector import ChargingConnector
from carconnectivity.observable import Observable
from carconnectivity.vehicle import ElectricVehicle

from carconnectivity_plugins.homekit.accessories.generic_accessory import GenericAccessory

if TYPE_CHECKING:
    from typing import Optional, Any

    from pyhap.service import Service
    from pyhap.accessory_driver import AccessoryDriver

    from carconnectivity.vehicle import GenericVehicle

    from carconnectivity_plugins.homekit.accessories.bridge import CarConnectivityBridge


LOG: logging.Logger = logging.getLogger("carconnectivity.plugins.homekit.charging_plug")


class ChargingPlugAccessory(GenericAccessory):
    """Charging Plug Accessory"""
    category: int = CATEGORY_SENSOR

    # pylint: disable-next=too-many-arguments,too-many-positional-arguments
    def __init__(self, driver: AccessoryDriver, bridge: CarConnectivityBridge, aid: int, id_str: str, vin: str, display_name: str,
                 vehicle: GenericVehicle) -> None:
        super().__init__(driver=driver, bridge=bridge, display_name=display_name, aid=aid, vin=vin, id_str=id_str)
        self.vehicle: GenericVehicle = vehicle

        # pyright: ignore[reportArgumentType]
        self.service: Optional[Service] = self.add_preload_service(service='ContactSensor',  # pyright: ignore[reportArgumentType]
                                                                   chars=['Name', 'ConfiguredName',  # pyright: ignore[reportArgumentType]
                                                                          'ContactSensorState', 'StatusFault'])

        self.char_contact_sensor_state: Optional[Characteristic] = None

        self.add_name_characteristics()
        self.add_status_fault_characteristic()

        if self.vehicle is not None and isinstance(self.vehicle, ElectricVehicle) and self.vehicle.charging is not None \
                and self.vehicle.charging.connector is not None:
            if self.vehicle.charging.connector.connection_state is not None and self.vehicle.charging.connector.connection_state.enabled:
                self.vehicle.charging.connector.connection_state.add_observer(self.__on_cc_connection_state_change, flag=Observable.ObserverEvent.VALUE_CHANGED)
                self.char_contact_sensor_state = self.service.configure_char('ContactSensorState')
                self.__on_cc_connection_state_change(self.vehicle.charging.connector.connection_state, Observable.ObserverEvent.VALUE_CHANGED)

    def __on_cc_connection_state_change(self, element: Any, flags: Observable.ObserverEvent) -> None:
        if flags & Observable.ObserverEvent.VALUE_CHANGED:
            if self.char_contact_sensor_state is not None:
                if element.value is None:
                    self.char_contact_sensor_state.set_value(0)
                    if self.char_status_fault is not None:
                        self.char_status_fault.set_value(0)
                if isinstance(element.value, ChargingConnector.ChargingConnectorConnectionState):
                    if element.value == ChargingConnector.ChargingConnectorConnectionState.CONNECTED:
                        self.char_contact_sensor_state.set_value(0)
                        if self.char_status_fault is not None:
                            self.char_status_fault.set_value(0)
                    elif element.value in [ChargingConnector.ChargingConnectorConnectionState.DISCONNECTED,
                                           ChargingConnector.ChargingConnectorConnectionState.UNSUPPORTED]:
                        self.char_contact_sensor_state.set_value(1)
                        if self.char_status_fault is not None:
                            self.char_status_fault.set_value(0)
                    else:
                        self.char_contact_sensor_state.set_value(1)
                        if self.char_status_fault is not None:
                            self.char_status_fault.set_value(1)
            else:
                LOG.debug('Unsupported event %s', flags)



""" HomeKit Charging Accessory """
from __future__ import annotations
from typing import TYPE_CHECKING

import threading
from datetime import datetime, timezone
import logging


from pyhap.characteristic import Characteristic
from pyhap.const import CATEGORY_OUTLET

from carconnectivity.errors import SetterError
from carconnectivity.commands import GenericCommand
from carconnectivity.command_impl import ChargingStartStopCommand
from carconnectivity.charging_connector import ChargingConnector
from carconnectivity.observable import Observable
from carconnectivity.vehicle import ElectricVehicle
from carconnectivity.charging import Charging
from carconnectivity.attributes import PowerAttribute
from carconnectivity.units import Power

from carconnectivity_plugins.homekit.accessories.generic_accessory import BatteryGenericVehicleAccessory

if TYPE_CHECKING:
    from typing import Optional, Any, Dict

    from pyhap.service import Service
    from pyhap.accessory_driver import AccessoryDriver

    from carconnectivity.vehicle import GenericVehicle

    from carconnectivity_plugins.homekit.accessories.bridge import CarConnectivityBridge


LOG: logging.Logger = logging.getLogger("carconnectivity.plugins.homekit.charging")


class ChargingAccessory(BatteryGenericVehicleAccessory):  # pylint: disable=too-many-instance-attributes
    """Charging Accessory"""

    category: int = CATEGORY_OUTLET

    # pylint: disable-next=too-many-arguments,too-many-positional-arguments
    def __init__(self, driver: AccessoryDriver, bridge: CarConnectivityBridge, aid: int, id_str: str, vin: str, display_name: str,
                 vehicle: GenericVehicle) -> None:
        super().__init__(driver=driver, bridge=bridge, display_name=display_name, aid=aid, vin=vin, id_str=id_str, vehicle=vehicle)

        self.update_remaining_duration_timer: Optional[threading.Timer] = None

        # pyright: ignore[reportArgumentType]
        self.service: Optional[Service] = self.add_preload_service(service='Outlet',  # pyright: ignore[reportArgumentType]
                                                                   chars=['Name', 'ConfiguredName',  # pyright: ignore[reportArgumentType]
                                                                          'On', 'OutletInUse', 'RemainingDuration', 'Consumption', 'StatusFault'])

        self.char_on: Optional[Characteristic] = None
        self.char_remaining_duration: Optional[Characteristic] = None
        self.char_consumption: Optional[Characteristic] = None
        self.char_outlet_in_use: Optional[Characteristic] = None

        self.charging_start_stop_command: Optional[GenericCommand] = None

        self.add_name_characteristics()
        self.add_status_fault_characteristic()
        self.add_soc_characteristic()

        if self.vehicle is not None and isinstance(self.vehicle, ElectricVehicle) and self.vehicle.charging is not None:
            if self.vehicle.charging.state is not None and self.vehicle.charging.state.enabled:
                self.vehicle.charging.state.add_observer(self.__on_cc_charging_state_change, flag=Observable.ObserverEvent.VALUE_CHANGED)
                self.char_on = self.service.configure_char('On', setter_callback=self.__on_hk_on_change)
                self.__on_cc_charging_state_change(self.vehicle.charging.state, Observable.ObserverEvent.VALUE_CHANGED)
                if self.vehicle.charging.commands is not None and self.vehicle.charging.commands.contains_command('start-stop'):
                    self.charging_start_stop_command = self.vehicle.charging.commands.commands['start-stop']

            if self.vehicle.charging.estimated_date_reached is not None and self.vehicle.charging.estimated_date_reached.enabled:
                self.vehicle.charging.estimated_date_reached.add_observer(self.__on_cc_estimated_date_reached_change,
                                                                          flag=Observable.ObserverEvent.UPDATED_NEW_MEASUREMENT)
                self.char_remaining_duration = self.service.configure_char('RemainingDuration')
                self.__on_cc_estimated_date_reached_change(self.vehicle.charging.estimated_date_reached,
                                                           Observable.ObserverEvent.UPDATED_NEW_MEASUREMENT)

            if self.vehicle.charging.power is not None and self.vehicle.charging.power.enabled:
                self.vehicle.charging.power.add_observer(self.__on_cc_power_change, flag=Observable.ObserverEvent.VALUE_CHANGED)
                self.char_consumption = self.service.configure_char('Consumption')
                self.__on_cc_power_change(self.vehicle.charging.power, Observable.ObserverEvent.VALUE_CHANGED)

            if self.vehicle.charging.connector is not None and self.vehicle.charging.connector.connection_state is not None \
                    and self.vehicle.charging.connector.connection_state.enabled:
                self.vehicle.charging.connector.connection_state.add_observer(self.__on_cc_connector_state_change,
                                                                              flag=Observable.ObserverEvent.VALUE_CHANGED)
                self.char_outlet_in_use = self.service.configure_char('OutletInUse')
                self.__on_cc_connector_state_change(self.vehicle.charging.connector.connection_state,
                                                    Observable.ObserverEvent.VALUE_CHANGED)

    def __on_cc_charging_state_change(self, element: Any, flags: Observable.ObserverEvent) -> None:
        if flags & Observable.ObserverEvent.VALUE_CHANGED:
            if self.char_on is not None:
                if element.value is None:
                    self.char_on.set_value(0)
                elif element.value in (Charging.ChargingState.OFF,
                                       Charging.ChargingState.READY_FOR_CHARGING):
                    self.char_on.set_value(0)
                elif element.value in (Charging.ChargingState.CHARGING,
                                       Charging.ChargingState.DISCHARGING,
                                       Charging.ChargingState.CONSERVATION):
                    self.char_on.set_value(1)
                elif element.value in (Charging.ChargingState.ERROR,
                                       Charging.ChargingState.UNSUPPORTED):
                    self.char_on.set_value(0)
                else:
                    self.char_on.set_value(0)
                    LOG.warning('unsupported chargingState: %s', element.value.value)
        else:
            LOG.debug('Unsupported event %s', flags)

    def __on_hk_on_change(self, value: Any) -> None:
        try:
            if self.charging_start_stop_command is not None and self.charging_start_stop_command.enabled:
                if value in [1, 2, 3]:
                    LOG.info('Switch charging on')
                    command_args: Dict[str, Any] = {}
                    command_args['command'] = ChargingStartStopCommand.Command.START
                    self.charging_start_stop_command.value = command_args
                elif value == 0:
                    LOG.info('Switch charging off')
                    command_args: Dict[str, Any] = {}
                    command_args['command'] = ChargingStartStopCommand.Command.STOP
                    self.charging_start_stop_command.value = command_args
                else:
                    LOG.error('Input for charging not understood: %d', value)
        except SetterError as setter_error:
            LOG.error('Error starting/stopping charging: %s', setter_error)
            self.set_status_fault(1, timeout=120)

    def __on_cc_estimated_date_reached_change(self, element: Any, flags: Observable.ObserverEvent) -> None:
        if flags & Observable.ObserverEvent.UPDATED_NEW_MEASUREMENT:
            if self.char_remaining_duration is not None:
                if element.enabled and element.value is not None and isinstance(element.value, datetime):
                    self.estimated_date_reached = element.value
                    self.__update_remaining_duration()
                    LOG.debug('Charging estimated date reached Changed: %s', self.estimated_date_reached.isoformat())
                else:
                    self.estimated_date_reached = None
                    self.char_remaining_duration.set_value(0)
                    LOG.debug('Charging estimated date reached Changed: None')
        else:
            LOG.debug('Unsupported event %s', flags)

    # pylint: disable=duplicate-code
    def __update_remaining_duration(self) -> None:
        if self.update_remaining_duration_timer is not None:
            self.update_remaining_duration_timer.cancel()
            self.update_remaining_duration_timer = None
        if self.char_remaining_duration is not None:
            remaining_duration: int = 0
            utc_now: datetime = datetime.now(tz=timezone.utc)
            if self.estimated_date_reached is not None and self.estimated_date_reached > utc_now:
                remaining_duration = round((self.estimated_date_reached - utc_now).total_seconds())
            self.char_remaining_duration.set_value(remaining_duration)
            if remaining_duration > 0 and not self.driver.stop_event.is_set():
                self.update_remaining_duration_timer = threading.Timer(interval=5.0, function=self.__update_remaining_duration)
                self.update_remaining_duration_timer.start()
    # pylint: enable=duplicate-code

    def __on_cc_power_change(self, element: Any, flags: Observable.ObserverEvent) -> None:
        if flags & Observable.ObserverEvent.VALUE_CHANGED:
            if self.char_consumption is not None:
                if isinstance(element, PowerAttribute) and element.value is not None:
                    self.char_consumption.set_value(element.power_in(unit=Power.W))
                    LOG.debug('Charging power Changed: %dW', element.power_in(unit=Power.W))
                else:
                    self.char_consumption.set_value(0)
        else:
            LOG.debug('Unsupported event %s', flags)

    def __on_cc_connector_state_change(self, element: Any, flags: Observable.ObserverEvent) -> None:
        if flags & Observable.ObserverEvent.VALUE_CHANGED:
            if self.char_outlet_in_use is not None:
                if element.value is None:
                    self.char_outlet_in_use.set_value(False)
                elif element.value == ChargingConnector.ChargingConnectorConnectionState.CONNECTED:
                    self.char_outlet_in_use.set_value(True)
                elif element.value in (ChargingConnector.ChargingConnectorConnectionState.DISCONNECTED,
                                       ChargingConnector.ChargingConnectorConnectionState.INVALID,
                                       ChargingConnector.ChargingConnectorConnectionState.UNSUPPORTED):
                    self.char_outlet_in_use.set_value(False)
                else:
                    self.char_outlet_in_use.set_value(False)
                    LOG.warning('unsupported charging connector state: %s', element.value.value)
        else:
            LOG.debug('Unsupported event %s', flags)



"""Module containing the commandline interface for the carconnectivity package."""
from __future__ import annotations
from typing import TYPE_CHECKING

import logging

from carconnectivity.carconnectivity_base import CLI

from carconnectivity_plugins.homekit._version import __version__

if TYPE_CHECKING:
    pass

LOG = logging.getLogger("carconnectivity-homekit")


def main() -> None:
    """
    Entry point for the car connectivity homekit application.

    This function initializes and starts the command-line interface (CLI) for the
    car connectivity application using the specified logger and application name.
    """
    cli: CLI = CLI(logger=LOG, name='carconnectivity-homekit', description='Commandline Interface to interact with Car Services of various brands',
                   subversion=__version__)
    cli.main()



""" HomeKit Bridge for CarConnectivity """
from __future__ import annotations
from typing import TYPE_CHECKING

import logging
import json

from pyhap.accessory import Bridge
from pyhap.accessory_driver import AccessoryDriver

from carconnectivity.observable import Observable
from carconnectivity.vehicle import GenericVehicle, ElectricVehicle
from carconnectivity._version import __version__ as __carconnectivity_version__

from carconnectivity_plugins.homekit.accessories.dummy_accessory import DummyAccessory
from carconnectivity_plugins.homekit.accessories.climatization import ClimatizationAccessory
from carconnectivity_plugins.homekit.accessories.charging import ChargingAccessory
from carconnectivity_plugins.homekit.accessories.charging_plug import ChargingPlugAccessory
from carconnectivity_plugins.homekit.accessories.outside_temperature import OutsideTemperatureAccessory
from carconnectivity_plugins.homekit.accessories.flashing import FlashingLightAccessory
from carconnectivity_plugins.homekit.accessories.locking_system import LockingAccessory
from carconnectivity_plugins.homekit._version import __version__


if TYPE_CHECKING:
    from typing import Optional, Any, Dict, List

    from carconnectivity.carconnectivity import CarConnectivity

LOG: logging.Logger = logging.getLogger("carconnectivity.plugins.homekit.bridge")


class CarConnectivityBridge(Bridge):
    """VWsfriend Bridge"""

    # pylint: disable-next=too-many-arguments,too-many-positional-arguments
    def __init__(self, car_connectivity: CarConnectivity, driver: AccessoryDriver, display_name: str = 'CarConnectivity',
                 accessory_config_file: str = '~/.carconnectivity/homekit-accessory.config', ignore_vins: Optional[List[str]] = None,
                 ignore_accessory_types: Optional[List[str]] = None) -> None:
        super().__init__(driver=driver, display_name=display_name, )

        self.set_info_service(f'{__carconnectivity_version__} (HomeKit Plugin {__version__})', 'Till Steinbach', 'CarConnectivity', None)

        self.car_connectivity: CarConnectivity = car_connectivity
        self.ignore_vins: List[str] = ignore_vins or []
        self.ignore_accessory_types: List[str] = ignore_accessory_types or []

        self.driver: AccessoryDriver = driver

        self.accessory_config_file: str = accessory_config_file
        self.__accessory_config: Dict[str, Dict[str, Any]] = {}
        self.next_aid: int = 100
        try:
            self.read_config()
        except FileNotFoundError:
            pass

        for identifier, accessory in self.__accessory_config.items():
            if 'ConfiguredName' in accessory:
                display_name = accessory['ConfiguredName']
            else:
                display_name = identifier
            placeholder_accessory = DummyAccessory(driver=driver, display_name=display_name, aid=accessory['aid'])
            if 'category' in accessory:
                placeholder_accessory.category = accessory['category']
            if 'services' in accessory:
                for service in accessory['services']:
                    placeholder_accessory.add_preload_service(service, chars=None)
            self.add_accessory(placeholder_accessory)

    def install_observers(self) -> None:
        """
        Installs observers for the car connectivity components.

        This method checks if the garage component of the car connectivity is not None.
        If it is not None, it adds an observer to the garage component to listen for updates.
        The observer is set to trigger on the ENABLED event and will execute at the end of a transaction.

        Returns:
            None
        """
        if self.car_connectivity.garage is not None:
            self.car_connectivity.garage.add_observer(observer=self.__on_garage_update, flag=Observable.ObserverEvent.ENABLED, on_transaction_end=True)

    def __on_garage_update(self, element: Any, flags: Observable.ObserverEvent) -> None:
        """Update the accessories when the garage is updated."""
        if (flags & (Observable.ObserverEvent.ENABLED)) and isinstance(element, GenericVehicle):
            self.update(vehicle=element)

    def persist_config(self) -> None:
        """
        Persist the accessory configuration to a file.

        This method writes the accessory configuration to the file specified by
        `self.accessory_config_file` in JSON format. If the file cannot be written,
        an error message is logged.

        Raises:
            ValueError: If there is an error during the JSON serialization process.
        """
        if self.accessory_config_file:
            try:
                with open(file=self.accessory_config_file, mode='w', encoding='utf8') as file:
                    json.dump(self.__accessory_config, fp=file)
                LOG.info('Writing accessory config file %s', self.accessory_config_file)
            except ValueError as err:
                LOG.info('Could not write homekit accessoryConfigFile %s (%s)', self.accessory_config_file, err)

    def read_config(self):
        """
        Reads the accessory configuration from a JSON file and updates the accessory configuration attribute.

        This method opens the accessory configuration file specified by `self.accessory_config_file`, reads its contents,
        and loads it as a JSON object into `self.__accessory_config`. It also logs the action of reading the configuration file.
        Additionally, it iterates through the accessory configurations and updates `self.next_aid` to ensure it is set to
        one more than the highest 'aid' value found in the configurations.

        Raises:
            FileNotFoundError: If the accessory configuration file does not exist.
            json.JSONDecodeError: If the file contents cannot be decoded as JSON.
        """
        with open(file=self.accessory_config_file, mode='r', encoding='utf8') as file:
            self.__accessory_config = json.load(fp=file)
            LOG.info('Reading homekit accessory config file %s', self.accessory_config_file)
            # Find the highest aid in the config and adjust next_aid accordingly
            for accessory_config in self.__accessory_config.values():
                if 'aid' in accessory_config and (accessory_config['aid'] + 1) > self.next_aid:
                    self.next_aid = accessory_config['aid'] + 1

    def update(self, vehicle: GenericVehicle):  # noqa: C901 # pylint: disable=too-many-branches,too-many-statements,too-many-locals
        """
        Update the bridge with the given vehicle's information and accessories.

        Args:
            vehicle (GenericVehicle): The vehicle object containing various attributes and statuses.

        This method performs the following steps:
        1. Checks if the vehicle and its VIN are enabled and not None.
        2. Retrieves and sets the vehicle's manufacturer, name, model, and software version.
        3. Updates or adds the following accessories to the bridge if they are enabled and not ignored:
            - ClimatizationAccessory
            - ChargingAccessory (for ElectricVehicle)
            - ChargingPlugAccessory (for ElectricVehicle)
            - OutsideTemperatureAccessory
            - FlashingLightAccessory
            - LockingAccessory
        4. If any configuration changes are made, updates the driver and persists the configuration.

        Note:
            - Accessories are added to the bridge if they are not already known.
            - If an accessory is known but not of the correct type, it is replaced.
            - The method logs debug information and persists configuration changes if any updates are made.
        """
        config_changed = False
        if vehicle.enabled and vehicle.vin.enabled and vehicle.vin.value is not None:
            vin: str = vehicle.vin.value
            if vin in self.ignore_vins:
                LOG.debug('Ignoring vehicle with VIN %s due to configuration', vin)
                return
            if vehicle.manufacturer.enabled and vehicle.manufacturer.value is not None:
                manufacturer: str = vehicle.manufacturer.value
            else:
                manufacturer: str = 'Unknown'
            if vehicle.name.enabled and vehicle.name.value is not None and len(vehicle.name.value) > 0:
                name: str = vehicle.name.value
            else:
                name: str = vin
            if vehicle.model.enabled and vehicle.model.value is not None and len(vehicle.model.value) > 0:
                model: str = vehicle.model.value
            else:
                model: str = 'Unknown'
            if vehicle.software is not None and vehicle.software.enabled and vehicle.software.version is not None and vehicle.software.version.enabled:
                vehicle_software_version: Optional[str] = vehicle.software.version.value
            else:
                vehicle_software_version: Optional[str] = None
            # Climatization
            climatization_aid: Optional[int] = self.get_existing_aid('Climatization', vin)
            # pylint: disable-next=too-many-boolean-expressions
            if 'Climatization' not in self.ignore_accessory_types \
                    and vehicle.climatization is not None and vehicle.climatization.enabled \
                    and (climatization_aid is None
                         or climatization_aid not in self.accessories
                         or not isinstance(self.accessories[climatization_aid], ClimatizationAccessory)):
                climatization_accessory = ClimatizationAccessory(driver=self.driver, bridge=self, aid=self.select_aid('Climatization', vin),
                                                                 id_str='Climatization', vin=vin, display_name=f'{name} Climatization',
                                                                 vehicle=vehicle)
                climatization_accessory.set_info_service(firmware_revision=vehicle_software_version, manufacturer=manufacturer, model=model,
                                                         serial_number=f'{vin}-climatization')
                self.set_config_item(climatization_accessory.id_str, climatization_accessory.vin, 'category', climatization_accessory.category)
                self.set_config_item(climatization_accessory.id_str, climatization_accessory.vin, 'services',
                                     [service.display_name for service in climatization_accessory.services])
                # Add the accessory to the bridge if not known
                if climatization_accessory.aid not in self.accessories:
                    self.add_accessory(climatization_accessory)
                # Replace the accessory if it is known but not of the correct type (was Dummy before)
                else:
                    self.accessories[climatization_accessory.aid] = climatization_accessory
                config_changed = True

            # Charging
            charging_aid: Optional[int] = self.get_existing_aid('Charging', vin)
            # pylint: disable-next=too-many-boolean-expressions
            if 'Charging' not in self.ignore_accessory_types \
                    and isinstance(vehicle, ElectricVehicle) and vehicle.charging is not None and vehicle.charging.enabled \
                    and (charging_aid is None
                         or charging_aid not in self.accessories or not isinstance(self.accessories[charging_aid], ChargingAccessory)):
                charging_accessory = ChargingAccessory(driver=self.driver, bridge=self, aid=self.select_aid('Charging', vin),
                                                       id_str='Charging', vin=vin, display_name=f'{name} Charging', vehicle=vehicle)
                charging_accessory.set_info_service(firmware_revision=vehicle_software_version, manufacturer=manufacturer, model=model,
                                                    serial_number=f'{vin}-charging')
                self.set_config_item(charging_accessory.id_str, charging_accessory.vin, 'category', charging_accessory.category)
                self.set_config_item(charging_accessory.id_str, charging_accessory.vin, 'services',
                                     [service.display_name for service in charging_accessory.services])
                # Add the accessory to the bridge if not known
                if charging_accessory.aid not in self.accessories:
                    self.add_accessory(charging_accessory)
                # Replace the accessory if it is known but not of the correct type (was Dummy before)
                else:
                    self.accessories[charging_accessory.aid] = charging_accessory
                config_changed = True

            # ChargingPlug
            plug_aid: Optional[int] = self.get_existing_aid('ChargingPlug', vin)
            # pylint: disable-next=too-many-boolean-expressions
            if 'ChargingPlug' not in self.ignore_accessory_types \
                    and isinstance(vehicle, ElectricVehicle) and vehicle.charging is not None and vehicle.charging.enabled \
                    and (plug_aid is None or plug_aid not in self.accessories or not isinstance(self.accessories[plug_aid], ChargingPlugAccessory)):
                charging_plug_accessory = ChargingPlugAccessory(driver=self.driver, bridge=self, aid=self.select_aid('ChargingPlug', vin),
                                                                id_str='ChargingPlug', vin=vin, display_name=f'{name} Charging Plug', vehicle=vehicle)
                charging_plug_accessory.set_info_service(firmware_revision=vehicle_software_version, manufacturer=manufacturer, model=model,
                                                         serial_number=f'{vin}-charging-plug')
                self.set_config_item(charging_plug_accessory.id_str, charging_plug_accessory.vin, 'category', charging_plug_accessory.category)
                self.set_config_item(charging_plug_accessory.id_str, charging_plug_accessory.vin, 'services',
                                     [service.display_name for service in charging_plug_accessory.services])
                # Add the accessory to the bridge if not known
                if charging_plug_accessory.aid not in self.accessories:
                    self.add_accessory(charging_plug_accessory)
                # Replace the accessory if it is known but not of the correct type (was Dummy before)
                else:
                    self.accessories[charging_plug_accessory.aid] = charging_plug_accessory
                config_changed = True

            # OutsideTemperature
            outside_temperature_aid: Optional[int] = self.get_existing_aid('OutsideTemperature', vin)
            # pylint: disable-next=too-many-boolean-expressions
            if 'OutsideTemperature' not in self.ignore_accessory_types \
                    and vehicle.outside_temperature is not None and vehicle.outside_temperature.enabled \
                    and (outside_temperature_aid is None
                         or outside_temperature_aid not in self.accessories or not isinstance(self.accessories[outside_temperature_aid],
                                                                                              OutsideTemperatureAccessory)):
                outside_temperature_accessory = OutsideTemperatureAccessory(driver=self.driver, bridge=self, aid=self.select_aid('OutsideTemperature', vin),
                                                                            id_str='OutsideTemperature', vin=vin, display_name=f'{name} Outside Temperature',
                                                                            vehicle=vehicle)
                outside_temperature_accessory.set_info_service(firmware_revision=vehicle_software_version, manufacturer=manufacturer, model=model,
                                                               serial_number=f'{vin}-outside-temperature')
                self.set_config_item(outside_temperature_accessory.id_str, outside_temperature_accessory.vin, 'category',
                                     outside_temperature_accessory.category)
                self.set_config_item(outside_temperature_accessory.id_str, outside_temperature_accessory.vin, 'services',
                                     [service.display_name for service in outside_temperature_accessory.services])
                # Add the accessory to the bridge if not known
                if outside_temperature_accessory.aid not in self.accessories:
                    self.add_accessory(outside_temperature_accessory)
                # Replace the accessory if it is known but not of the correct type (was Dummy before)
                else:
                    self.accessories[outside_temperature_accessory.aid] = outside_temperature_accessory
                config_changed = True

            # FlashingAccessory
            flashing_light_aid: Optional[int] = self.get_existing_aid('FlashingLight', vin)
            # pylint: disable-next=too-many-boolean-expressions
            if 'FlashingLight' not in self.ignore_accessory_types \
                    and vehicle.commands is not None and vehicle.commands is not None and 'honk-flash' in vehicle.commands.commands \
                    and (flashing_light_aid is None or flashing_light_aid not in self.accessories or not isinstance(self.accessories[flashing_light_aid],
                                                                                                                    FlashingLightAccessory)):
                flashing_light_accessory = FlashingLightAccessory(driver=self.driver, bridge=self, aid=self.select_aid('FlashingLight', vin),
                                                                  id_str='FlashingLight', vin=vin, display_name=f'{name} Flashing',
                                                                  vehicle=vehicle)
                flashing_light_accessory.set_info_service(firmware_revision=vehicle_software_version, manufacturer=manufacturer, model=model,
                                                          serial_number=f'{vin}-flashing-light')
                self.set_config_item(flashing_light_accessory.id_str, flashing_light_accessory.vin, 'category',
                                     flashing_light_accessory.category)
                self.set_config_item(flashing_light_accessory.id_str, flashing_light_accessory.vin, 'services',
                                     [service.display_name for service in flashing_light_accessory.services])
                # Add the accessory to the bridge if not known
                if flashing_light_accessory.aid not in self.accessories:
                    self.add_accessory(flashing_light_accessory)
                # Replace the accessory if it is known but not of the correct type (was Dummy before)
                else:
                    self.accessories[flashing_light_accessory.aid] = flashing_light_accessory
                config_changed = True

            # LockingAccessory
            locking_aid: Optional[int] = self.get_existing_aid('Locking', vin)
            # pylint: disable-next=too-many-boolean-expressions
            if 'Locking' not in self.ignore_accessory_types \
                    and vehicle.doors is not None and vehicle.doors.commands is not None \
                    and 'lock-unlock' in vehicle.doors.commands.commands \
                    and (locking_aid is None or locking_aid not in self.accessories or not isinstance(self.accessories[locking_aid],
                                                                                                      LockingAccessory)):
                locking_accessory = LockingAccessory(driver=self.driver, bridge=self, aid=self.select_aid('Locking', vin),
                                                     id_str='Locking', vin=vin, display_name=f'{name} Locking',
                                                     vehicle=vehicle)
                locking_accessory.set_info_service(firmware_revision=vehicle_software_version, manufacturer=manufacturer, model=model,
                                                   serial_number=f'{vin}-locking')
                self.set_config_item(locking_accessory.id_str, locking_accessory.vin, 'category', locking_accessory.category)
                self.set_config_item(locking_accessory.id_str, locking_accessory.vin, 'services',
                                     [service.display_name for service in locking_accessory.services])
                # Add the accessory to the bridge if not known
                if locking_accessory.aid not in self.accessories:
                    self.add_accessory(locking_accessory)
                # Replace the accessory if it is known but not of the correct type (was Dummy before)
                else:
                    self.accessories[locking_accessory.aid] = locking_accessory
                config_changed = True

        if config_changed:
            self.driver.config_changed()
            LOG.debug('Config changed, updating driver and persisting config')

    def get_existing_aid(self, id_str: str, vin: str) -> Optional[int]:
        """
        Retrieve the existing accessory ID (AID) for a given identifier string and vehicle identification number (VIN).

        Args:
            id_str (str): The identifier string associated with the accessory.
            vin (str): The vehicle identification number.

        Returns:
            Optional[int]: The accessory ID (AID) if it exists, otherwise None.
        """
        identifier = f'{vin}-{id_str}'
        aid: Optional[int] = None
        if identifier in self.__accessory_config and 'aid' in self.__accessory_config[identifier] \
                and self.__accessory_config[identifier]['aid'] is not None:
            aid = self.__accessory_config[identifier]['aid']
        return aid

    def select_aid(self, id_str: str, vin: str) -> int:
        """
        Selects or generates an accessory ID (aid) for a given accessory identifier.

        This method checks if an accessory ID (aid) already exists for the given
        combination of `id` and `vin` in the accessory configuration. If it exists,
        it returns the existing aid. If it does not exist, it generates a new aid,
        updates the accessory configuration, and returns the new aid.

        Args:
            id_str (str): The unique identifier for the accessory type.
            vin (str): The vehicle identification number.

        Returns:
            int: The accessory ID (aid) for the given accessory identifier.
        """
        aid = self.get_existing_aid(id_str=id_str, vin=vin)
        identifier = f'{vin}-{id_str}'
        if aid is None:
            new_aid: int = self.next_aid
            self.next_aid += 1
            if identifier in self.__accessory_config:
                self.__accessory_config[identifier]['aid'] = new_aid
            else:
                self.__accessory_config[identifier] = {'aid': new_aid}
            return new_aid
        return aid

    def set_config_item(self, id_str: str, vin: str, config_key: str, item: Any) -> None:
        """
        Set a configuration item for a specific accessory identified by a combination of VIN and ID string.

        Args:
            id_str (str): The identifier string for the accessory.
            vin (str): The vehicle identification number.
            config_key (str): The key for the configuration item.
            item (Any): The value to be set for the configuration item.

        Returns:
            None
        """
        identifier: str = f'{vin}-{id_str}'
        if identifier in self.__accessory_config:
            self.__accessory_config[identifier][config_key] = item
        else:
            self.__accessory_config[identifier] = {config_key: item}
        self.next_aid += 1

    def get_config_item(self, id_str: str, vin: str, config_key: str) -> Optional[Any]:
        """
        Retrieve a configuration item for a specific accessory.

        Args:
            id_str (str): The identifier string for the accessory.
            vin (str): The vehicle identification number.
            config_key (str): The key for the specific configuration item.

        Returns:
            Optional[Any]: The value of the configuration item if found, otherwise None.
        """
        identifier: str = f'{vin}-{id_str}'
        if identifier in self.__accessory_config and config_key in self.__accessory_config[identifier]:
            return self.__accessory_config[identifier][config_key]
        return None



""" User interface for the HomeKit plugin in the Car Connectivity application. """
from __future__ import annotations
from asyncio import AbstractEventLoop
from typing import TYPE_CHECKING

import os
import asyncio
from io import BytesIO
from uuid import UUID
from pyqrcode import pyqrcode

import flask
from flask_wtf import FlaskForm
from wtforms import SubmitField
from flask_login import login_required

from carconnectivity_plugins.base.plugin import BasePlugin
from carconnectivity_plugins.base.ui.plugin_ui import BasePluginUI

from carconnectivity_plugins.homekit.plugin import Plugin

if TYPE_CHECKING:
    from typing import Optional, List, Dict, Union, Literal

    from carconnectivity.carconnectivity import CarConnectivity


class PluginUI(BasePluginUI):
    """
    A user interface class for the HomeKit plugin in the Car Connectivity application.
    """
    def __init__(self, plugin: BasePlugin):
        super().__init__(plugin)
        self.blueprint: Optional[flask.Blueprint] = flask.Blueprint(name='homekit', import_name='carconnectivity-plugin-homekit', url_prefix='/homekit',
                                                                    template_folder=os.path.dirname(__file__) + '/templates')

        @self.blueprint.route('/status', methods=['GET'])
        @login_required
        def status():
            return flask.render_template('plugins/status.html', current_app=flask.current_app)

        class HomekitForm(FlaskForm):
            """
            HomekitForm class for handling the Homekit unpairing form.

            Attributes:
                unpair (SubmitField): A submit button labeled 'Unpair' for initiating the unpairing process.
            """
            unpair = SubmitField('Unpair')

        @self.blueprint.route('/pairing', methods=['GET', 'POST'])
        @login_required
        def pairing():
            car_connectivity: Optional[CarConnectivity] = flask.current_app.extensions['car_connectivity']
            if car_connectivity is not None:
                if 'homekit' in car_connectivity.plugins.plugins and car_connectivity.plugins.plugins['homekit'] is not None \
                        and isinstance(car_connectivity.plugins.plugins['homekit'], Plugin):
                    plugin: Plugin = car_connectivity.plugins.plugins['homekit']

                    form = HomekitForm()

                    if form.unpair.data:
                        clients: list[UUID] = list(plugin._driver.state.paired_clients.keys()).copy()  # pylint: disable=protected-access
                        for client in clients:
                            try:
                                asyncio.get_event_loop()
                            except RuntimeError as ex:
                                if "There is no current event loop in thread" in str(ex):
                                    loop: AbstractEventLoop = asyncio.new_event_loop()
                                    asyncio.set_event_loop(loop)
                            plugin._driver.unpair(client)  # pylint: disable=protected-access
                        plugin._driver.config_changed()  # pylint: disable=protected-access
                        flask.flash('Unpaired the Homekit bridge. You can now pair again')

                    return flask.render_template('pairing.html', form=form, current_app=flask.current_app, homekit_plugin=plugin)
            return flask.abort(500, "HomeKit plugin not found or wrong structure.")

        @self.blueprint.route('/accessories', methods=['GET'])
        @login_required
        def accessories():
            car_connectivity: Optional[CarConnectivity] = flask.current_app.extensions['car_connectivity']
            if car_connectivity is not None:
                if 'homekit' in car_connectivity.plugins.plugins and car_connectivity.plugins.plugins['homekit'] is not None \
                        and isinstance(car_connectivity.plugins.plugins['homekit'], Plugin):
                    plugin: Plugin = car_connectivity.plugins.plugins['homekit']
                    return flask.render_template('accessories.html', current_app=flask.current_app, homekit_plugin=plugin)
            return flask.abort(500, "HomeKit plugin not found or wrong structure.")

        @self.blueprint.route('/homekit-qr.png', methods=['GET'])
        @login_required
        def homekit_qr():
            if 'car_connectivity' not in flask.current_app.extensions:
                flask.abort(500, "Status picture doesn't exist.")
            car_connectivity: Optional[CarConnectivity] = flask.current_app.extensions['car_connectivity']
            if car_connectivity is not None:
                if 'homekit' in car_connectivity.plugins.plugins and car_connectivity.plugins.plugins['homekit'] is not None \
                        and isinstance(car_connectivity.plugins.plugins['homekit'], Plugin):
                    plugin: Plugin = car_connectivity.plugins.plugins['homekit']
                    if (accessory := plugin._driver.accessory) is not None:  # pylint: disable=protected-access
                        xhm_uri = accessory.xhm_uri()
                        qrcode = pyqrcode.create(xhm_uri)
                        img_io = BytesIO()
                        qrcode.png(img_io, scale=12)
                        img_io.seek(0)
                        return flask.send_file(img_io, mimetype='image/png')
            return flask.abort(500, "HomeKit plugin not found or wrong structure.")

    def get_nav_items(self) -> List[Dict[Literal['text', 'url', 'sublinks', 'divider'], Union[str, List]]]:
        """
        Generates a list of navigation items for the HomeKit plugin UI.
        """
        return [{"text": "Pairing", "url": flask.url_for('plugins.homekit.pairing')},
                {"text": "Accessories", "url": flask.url_for('plugins.homekit.accessories')},]

    def get_title(self) -> str:
        """
        Returns the title of the plugin.

        Returns:
            str: The title of the plugin, which is "HomeKit".
        """
        return "HomeKit"



""" This module defines the GenericAccessory class. """
from __future__ import annotations
from typing import TYPE_CHECKING

import logging
import threading

from pyhap.accessory import Accessory
from pyhap.characteristic import Characteristic

from carconnectivity.observable import Observable
from carconnectivity.vehicle import ElectricVehicle
from carconnectivity.charging import Charging

if TYPE_CHECKING:
    from typing import Optional, Any

    from pyhap.accessory_driver import AccessoryDriver
    from pyhap.service import Service

    from carconnectivity.vehicle import GenericVehicle, ElectricDrive

    from carconnectivity_plugins.homekit.accessories.bridge import CarConnectivityBridge

LOG: logging.Logger = logging.getLogger("carconnectivity.plugins.homekit.generic_accessory")


class GenericAccessory(Accessory):  # pylint: disable=too-many-instance-attributes
    """
    GenericAccessory is a class that represents a generic accessory in a HomeKit environment.

    Attributes:
        driver (AccessoryDriver): The driver for the accessory.
        bridge (CarConnectivityBridge): The bridge for car connectivity.
        id_str (str): The identifier string for the accessory.
        vin (str): The vehicle identification number.
        display_name (str): The display name of the accessory.
        service (Optional[Service]): The service associated with the accessory.
        char_status_fault (Optional[Characteristic]): The characteristic for status fault.
        char_configured_name (Optional[Characteristic]): The characteristic for configured name.
        char_name (Optional[Characteristic]): The characteristic for name.
    """
    # pylint: disable-next=too-many-arguments,too-many-positional-arguments
    def __init__(self, driver: AccessoryDriver, bridge: CarConnectivityBridge, aid: int, id_str: str, vin: str, display_name: str) -> None:
        super().__init__(driver=driver, display_name=display_name, aid=aid)

        self.driver: AccessoryDriver = driver
        self.bridge: CarConnectivityBridge = bridge
        self.id_str: str = id_str
        self.vin: str = vin

        self.service: Optional[Service] = None

        self.char_status_fault: Optional[Characteristic] = None
        self.__char_status_fault_timer: Optional[threading.Timer] = None

        self.char_configured_name: Optional[Characteristic] = None
        self.char_name: Optional[Characteristic] = None

    def add_name_characteristics(self) -> None:
        """
        Adds name characteristics to the accessory's service.

        This method retrieves the configured name for the accessory from the bridge configuration.
        If no configured name is found, it defaults to the display name of the accessory.
        It then configures the 'ConfiguredName' and 'Name' characteristics for the accessory's service.

        Returns:
            None
        """
        configured_name: Optional[str] = self.bridge.get_config_item(self.id_str, self.vin, 'ConfiguredName')
        if configured_name is None:
            configured_name: Optional[str] = self.display_name
        if self.service is not None:
            self.char_configured_name: Optional[Characteristic] = self.service.configure_char(char_name='ConfiguredName', value=configured_name,
                                                                                              setter_callback=self.__on_configured_name_changed)
            self.char_name: Optional[Characteristic] = self.service.configure_char(char_name='Name', value=configured_name)

    def __on_configured_name_changed(self, value: Any) -> None:
        """
        Callback function that is triggered when the configured name changes.

        This function updates the configuration and persists it when the configured name changes.
        It also updates the characteristic values for `char_configured_name` and `char_name` if they are not None.

        Args:
            value (Any): The new configured name value.
        """
        if value is not None and len(value) > 0:
            self.bridge.set_config_item(id_str=self.id_str, vin=self.vin, config_key='ConfiguredName', item=value)
            self.bridge.persist_config()
            if self.char_configured_name is not None:
                self.char_configured_name.set_value(value)
            if self.char_name is not None:
                self.char_name.set_value(value)

    def add_status_fault_characteristic(self) -> None:
        """
        Adds a 'StatusFault' characteristic to the accessory's service.

        This method checks if the accessory's service is not None and then
        configures a 'StatusFault' characteristic with an initial value of 0.

        Returns:
            None
        """
        if self.service is not None:
            self.char_status_fault = self.service.configure_char(char_name='StatusFault', value=0)

    def set_status_fault(self, value, timeout: float = 0, timeout_value: Optional[int] = None) -> None:
        """
        Sets the status fault characteristic to the given value. If a timeout is specified, the status fault will be reset
        to the timeout_value after the timeout period.

        Args:
            value: The value to set for the status fault characteristic.
            timeout (float, optional): The time in seconds after which the status fault should be reset. Defaults to 0.
            timeout_value (Optional[int], optional): The value to reset the status fault to after the timeout. If not
                                                     specified, the current value of the status fault characteristic will
                                                     be used. Defaults to None.

        Returns:
            None
        """
        if self.char_status_fault is not None:
            if self.__char_status_fault_timer is not None and self.__char_status_fault_timer.is_alive():
                self.__char_status_fault_timer.cancel()
                self.__char_status_fault_timer = None
            if timeout == 0:
                self.char_status_fault.set_value(value)
            else:
                if timeout_value is None:
                    timeout_value = self.char_status_fault.get_value()
                self.char_status_fault.set_value(value)
                self.__char_status_fault_timer = threading.Timer(timeout, self.__reset_status_fault, [timeout_value])
                self.__char_status_fault_timer.daemon = True
                self.__char_status_fault_timer.start()

    def __reset_status_fault(self, value: int) -> None:
        """
        Resets the status fault characteristic to the given value.

        Args:
            value (int): The value to set for the status fault characteristic.

        Returns:
            None
        """
        if self.char_status_fault is not None:
            self.char_status_fault.set_value(value)


class BatteryGenericVehicleAccessory(GenericAccessory):
    """Vehcile Accessory with battery characteristics"""

    # pylint: disable-next=too-many-arguments,too-many-positional-arguments
    def __init__(self, driver: AccessoryDriver, bridge: CarConnectivityBridge, aid: int, id_str: str, vin: str, display_name: str,
                 vehicle: GenericVehicle) -> None:
        super().__init__(driver=driver, bridge=bridge, display_name=display_name, aid=aid, vin=vin, id_str=id_str)

        self.vehicle: GenericVehicle = vehicle

        self.battery_service: Optional[Service] = None
        self.char_battery_level: Optional[Characteristic] = None
        self.char_status_low_battery: Optional[Characteristic] = None
        self.char_charging_state: Optional[Characteristic] = None

    def add_soc_characteristic(self) -> None:
        """
        Adds the State of Charge (SoC) characteristic to the accessory if the vehicle is an electric vehicle.

        This method checks if the vehicle is an instance of `ElectricVehicle`. If so, it retrieves the electric drive
        and its level. If the level is enabled, it adds an observer to monitor changes in the level value. It then
        adds a battery service with characteristics for battery level, low battery status, and charging state.
        The battery level and low battery status are configured and set based on the current level value of the
        electric drive. If the accessory has a service, the battery service is linked to it.

        Returns:
            None
        """
        if self.battery_service is None and isinstance(self.vehicle, ElectricVehicle):
            electric_drive: ElectricDrive = self.vehicle.get_electric_drive()
            if electric_drive is not None and electric_drive.level is not None and electric_drive.level.enabled:
                electric_drive.level.add_observer(self._on_level_change, Observable.ObserverEvent.VALUE_CHANGED)
                self.battery_service: Optional[Service] = self.add_preload_service(service='BatteryService',  # pyright: ignore[reportArgumentType]
                                                                                   chars=['BatteryLevel',  # pyright: ignore[reportArgumentType]
                                                                                          'StatusLowBattery',
                                                                                          'ChargingState'])
                self.char_battery_level = self.battery_service.configure_char('BatteryLevel')
                self.char_battery_level.set_value(electric_drive.level.value)
                self.char_status_low_battery = self.battery_service.configure_char('StatusLowBattery')
                self._set_low_battery_status(electric_drive.level.value)
                if self.service is not None:
                    self.service.add_linked_service(self.battery_service)
                if self.vehicle.charging is not None and self.vehicle.charging.state is not None and self.vehicle.charging.state.enabled:
                    self.vehicle.charging.state.add_observer(self._on_charging_state, Observable.ObserverEvent.VALUE_CHANGED)
                    self.char_charging_state = self.battery_service.configure_char('ChargingState')
                    self._set_charging_state(self.vehicle.charging.state.value)

    def _set_low_battery_status(self, level: Optional[float]) -> None:
        if self.char_status_low_battery is not None:
            if level is None or level > 10:
                self.char_status_low_battery.set_value(0)
            else:
                self.char_status_low_battery.set_value(1)

    def _set_charging_state(self, charging_state) -> None:
        if self.char_charging_state is not None:
            if charging_state is not None:
                if charging_state in (Charging.ChargingState.OFF,
                                      Charging.ChargingState.READY_FOR_CHARGING,
                                      Charging.ChargingState.UNSUPPORTED):
                    self.char_charging_state.set_value(0)
                elif charging_state in (Charging.ChargingState.CHARGING,
                                        Charging.ChargingState.DISCHARGING,
                                        Charging.ChargingState.CONSERVATION):
                    self.char_charging_state.set_value(1)
                elif charging_state == Charging.ChargingState.ERROR:
                    self.char_charging_state.set_value(2)
                else:
                    self.char_charging_state.set_value(2)
                    LOG.warning('unsupported charging state: %s', charging_state)
            else:
                self.char_charging_state.set_value(2)

    def _on_level_change(self, element: Any, flags: Observable.ObserverEvent) -> None:
        if flags & Observable.ObserverEvent.VALUE_CHANGED:
            if element.value is not None and self.char_battery_level is not None:
                self.char_battery_level.set_value(element.value)
                self._set_low_battery_status(element.value)

    def _on_charging_state(self, element: Any, flags: Observable.ObserverEvent) -> None:
        if flags & Observable.ObserverEvent.VALUE_CHANGED:
            if element.value is not None and self.char_charging_state is not None:
                self._set_charging_state(element.value)



""" HomeKit Locking Accessory """
from __future__ import annotations
from typing import TYPE_CHECKING

import logging

from pyhap.characteristic import Characteristic
from pyhap.const import CATEGORY_DOOR_LOCK

from carconnectivity.errors import SetterError
from carconnectivity.attributes import Observable
from carconnectivity.commands import GenericCommand
from carconnectivity.command_impl import LockUnlockCommand
from carconnectivity.doors import Doors

from carconnectivity_plugins.homekit.accessories.generic_accessory import GenericAccessory

if TYPE_CHECKING:
    from typing import Optional, Any, Dict

    from pyhap.service import Service
    from pyhap.accessory_driver import AccessoryDriver

    from carconnectivity.vehicle import GenericVehicle

    from carconnectivity_plugins.homekit.accessories.bridge import CarConnectivityBridge


LOG: logging.Logger = logging.getLogger("carconnectivity.plugins.homekit.locking")


class LockingAccessory(GenericAccessory):
    """Flashing Light Accessory"""
    category: int = CATEGORY_DOOR_LOCK

    # pylint: disable-next=too-many-arguments,too-many-positional-arguments
    def __init__(self, driver: AccessoryDriver, bridge: CarConnectivityBridge, aid: int, id_str: str, vin: str, display_name: str,
                 vehicle: GenericVehicle) -> None:
        super().__init__(driver=driver, bridge=bridge, display_name=display_name, aid=aid, vin=vin, id_str=id_str)
        self.vehicle: GenericVehicle = vehicle

        # pyright: ignore[reportArgumentType]
        self.service: Optional[Service] = self.add_preload_service(service='LockMechanism',  # pyright: ignore[reportArgumentType]
                                                                   chars=['Name', 'ConfiguredName',  # pyright: ignore[reportArgumentType]
                                                                          'LockCurrentState', 'LockTargetState', 'StatusFault'])

        self.char_lock_current_state: Optional[Characteristic] = None
        self.char_lock_target_state: Optional[Characteristic] = None

        self.add_name_characteristics()
        self.add_status_fault_characteristic()

        if self.vehicle is not None and self.vehicle.doors is not None:
            if self.vehicle.doors.commands is not None and self.vehicle.commands.commands is not None \
                    and self.vehicle.doors.commands.contains_command('lock-unlock'):
                self.lock_unlock_command: GenericCommand = self.vehicle.doors.commands.commands['lock-unlock']
                self.char_lock_target_state = self.service.configure_char('LockTargetState', setter_callback=self.__on_hk_lock_target_state_change)
                self.char_lock_target_state.allow_invalid_client_values = True
            if self.vehicle.doors.lock_state is not None and self.vehicle.doors.lock_state.enabled:
                self.vehicle.doors.lock_state.add_observer(self.__on_cc_lock_state_change, flag=Observable.ObserverEvent.VALUE_CHANGED)
                self.char_lock_current_state = self.service.configure_char('LockCurrentState')
                self.__on_cc_lock_state_change(self.vehicle.doors.lock_state, flags=Observable.ObserverEvent.VALUE_CHANGED)

    def __on_hk_lock_target_state_change(self, value: Any) -> None:
        if self.char_lock_target_state is not None:
            if self.lock_unlock_command is not None and self.lock_unlock_command.enabled:
                command_args: Dict[str, Any] = {}
                try:
                    if value == 1:
                        LOG.info('Lock car')
                        command_args['command'] = LockUnlockCommand.Command.LOCK
                        self.lock_unlock_command.value = command_args
                    elif value == 0:
                        LOG.info('Unlock car')
                        command_args['command'] = LockUnlockCommand.Command.UNLOCK
                        self.lock_unlock_command.value = command_args
                    else:
                        LOG.error('Input for lock target not understood: %d', value)
                        self.set_status_fault(1, timeout=120)
                except SetterError as setter_error:
                    LOG.error('Error locking or unlocking: %s', setter_error)
                    if self.char_lock_current_state is not None and self.char_lock_current_state.value in [1, 3]:
                        self.char_lock_target_state.set_value(1)
                    else:
                        self.char_lock_target_state.set_value(0)
                    self.set_status_fault(1, timeout=120)
            else:
                LOG.error('Locking cannot be controlled')
                if self.char_lock_current_state is not None and self.char_lock_current_state.value in [1, 3]:
                    self.char_lock_target_state.set_value(1)
                else:
                    self.char_lock_target_state.set_value(0)
                self.set_status_fault(1, timeout=120)

    def __on_cc_lock_state_change(self, element: Any, flags: Observable.ObserverEvent) -> None:
        if flags & Observable.ObserverEvent.VALUE_CHANGED:
            if self.char_lock_current_state is not None:
                if element.value == Doors.LockState.LOCKED:
                    self.char_lock_current_state.set_value(1)
                    if self.char_lock_target_state is not None:
                        self.char_lock_target_state.set_value(1)
                elif element.value == Doors.LockState.UNLOCKED:
                    self.char_lock_current_state.set_value(0)
                    if self.char_lock_target_state is not None:
                        self.char_lock_target_state.set_value(0)
                elif element.value == Doors.LockState.INVALID:
                    self.char_lock_current_state.set_value(3)
                    if self.char_lock_target_state is not None:
                        self.char_lock_target_state.set_value(1)
                else:
                    self.char_lock_current_state.set_value(3)
                    if self.char_lock_target_state is not None:
                        self.char_lock_target_state.set_value(1)
                    LOG.warning('unsupported lock state: %s', element.value)



""" Definitions for custom characteristics """

CUSTOM_CHARACTERISTICS = {
    "Consumption": {
        "Format": "uint32",
        "Permissions": [
            "pr",
            "ev"
        ],
        "UUID": "E863F10D-079E-48FF-8F27-9C2605A29F52",
        "maxValue": 4294967295,
        "minStep": 1,
        "minValue": 0,
        "unit": "watt"
    }
}



"""Module implements the plugin to connect with Homekit"""
from __future__ import annotations
from typing import TYPE_CHECKING

import os
import re
import threading
import logging
from pathlib import Path

from pyhap.accessory_driver import AccessoryDriver

from carconnectivity.errors import ConfigurationError
from carconnectivity.util import config_remove_credentials
from carconnectivity_plugins.base.plugin import BasePlugin
from carconnectivity_plugins.homekit._version import __version__

from carconnectivity_plugins.homekit.accessories.custom_characteristics import CUSTOM_CHARACTERISTICS
from carconnectivity_plugins.homekit.accessories.bridge import CarConnectivityBridge

if TYPE_CHECKING:
    from typing import Dict, Optional, List
    from carconnectivity.carconnectivity import CarConnectivity

LOG: logging.Logger = logging.getLogger("carconnectivity.plugins.homekit")


class Plugin(BasePlugin):
    """
    Plugin class for HomeKit integration.
    Args:
        car_connectivity (CarConnectivity): An instance of CarConnectivity.
        config (Dict): Configuration dictionary containing connection details.
    """
    def __init__(self, plugin_id: str, car_connectivity: CarConnectivity, config: Dict) -> None:  # pylint: disable=too-many-branches, too-many-statements
        BasePlugin.__init__(self, plugin_id=plugin_id, car_connectivity=car_connectivity, config=config)

        self._background_thread: Optional[threading.Thread] = None
        self.stop_event = threading.Event()

        # Configure logging
        if 'log_level' in config and config['log_level'] is not None:
            config['log_level'] = config['log_level'].upper()
            if config['log_level'] in logging._nameToLevel:
                LOG.setLevel(config['log_level'])
                self.log_level._set_value(config['log_level'])  # pylint: disable=protected-access
            else:
                raise ConfigurationError(f'Invalid log level: "{config["log_level"]}" not in {list(logging._nameToLevel.keys())}')
        LOG.info("Loading homekit plugin with config %s", config_remove_credentials(self.config))

        if 'address' in config and config['address'] is not None:
            address: Optional[str] = config['address']
        else:
            address = None

        if 'port' in config and config['port'] is not None:
            port: int = config['port']
            if port > 65535 or port < 1:
                raise ConfigurationError(f'Invalid port: "{port}" not in range 1-65535')
        else:
            port = 51234

        if 'pincode' in config and config['pincode'] is not None:
            pincode: Optional[str] = config['pincode']
            if pincode is not None and not re.match(pattern=r'^\d{3}-\d{2}-\d{3}$', string=pincode):
                raise ConfigurationError(f'Invalid pincode format: "{pincode}". Expected format is "xxx-xx-xxx" where x is a digit.')
        else:
            pincode = None

        if 'accessory_state_file' in config and config['accessory_state_file'] is not None:
            accessory_state_file: str = os.path.expanduser(config['accessory_state_file'])
        else:
            accessory_state_file: str = os.path.expanduser('~/.carconnectivity/homekit-accessory.state')
        file = Path(accessory_state_file)
        file.parent.mkdir(parents=True, exist_ok=True)

        if 'accessory_config_file' in config and config['accessory_config_file'] is not None:
            accessory_config_file: str = os.path.expanduser(config['accessory_config_file'])
        else:
            accessory_config_file: str = os.path.expanduser('~/.carconnectivity/homekit-accessory.config')
        file = Path(accessory_config_file)
        file.parent.mkdir(parents=True, exist_ok=True)

        if 'ignore_vins' in config and config['ignore_vins'] is not None:
            ignore_vins: List[str] = config['ignore_vins']
        else:
            ignore_vins = []

        if 'ignore_accessory_types' in config and config['ignore_accessory_types'] is not None:
            ignore_accessory_types: List[str] = config['ignore_accessory_types']
        else:
            ignore_accessory_types = []

        # Add the accessory driver
        self._driver = AccessoryDriver(address=address, port=port, pincode=pincode, persist_file=accessory_state_file)

        for characteristic_key, characteristic in CUSTOM_CHARACTERISTICS.items():
            self._driver.loader.char_types[characteristic_key] = characteristic

        self._bridge = CarConnectivityBridge(driver=self._driver, car_connectivity=car_connectivity, accessory_config_file=accessory_config_file,
                                             ignore_vins=ignore_vins, ignore_accessory_types=ignore_accessory_types)
        self._driver.add_accessory(self._bridge)

    def startup(self) -> None:
        LOG.info("Starting Homekit plugin")
        self._background_thread = threading.Thread(target=self._driver.start, daemon=False)
        self._background_thread.start()
        update_thread = threading.Timer(interval=5.0, function=self.__delayed_update)
        update_thread.daemon = True
        update_thread.start()
        LOG.debug("Starting Homekit plugin done")

    def __delayed_update(self) -> None:
        self.stop_event.wait(5.0)
        if not self.stop_event.is_set():
            self._bridge.install_observers()
            if self.car_connectivity.garage is not None:
                for vehicle in self.car_connectivity.garage.list_vehicles():
                    self._bridge.update(vehicle=vehicle)

    def shutdown(self) -> None:
        self.stop_event.set()
        self._driver.stop()
        if self._background_thread is not None:
            self._background_thread.join()
        return super().shutdown()

    def get_version(self) -> str:
        return __version__

    def get_type(self) -> str:
        """
        Returns the type of the plugin.

        :return: A string representing the type of the plugin.
        """
        return "carconnectivity-plugin-homekit"
